------------------------------------------------------------------------------------------
-- HEIG-VD ///////////////////////////////////////////////////////////////////////////////
-- Haute Ecole d'Ingenerie et de Gestion du Canton de Vaud
-- School of Business and Engineering in Canton de Vaud
------------------------------------------------------------------------------------------
-- REDS Institute ////////////////////////////////////////////////////////////////////////
-- Reconfigurable Embedded Digital Systems
------------------------------------------------------------------------------------------
--
-- File                 : DE1_SOC_golden_top.vhd
-- Author               : Flavio Capitao (FCC)
-- Date                 : 13.04.2017
--
-- Context              : Tic-Tac-Toe game
--
------------------------------------------------------------------------------------------
-- Description : DE1-SoC golden top for TicTacToe game
--   
--               SW9   : choose master ('1',UP) or slave ('0',DOWN)
--               LEDR9 : effective master ('1',ON) or slave ('0',OFF)
--               LEDR8 : slave select line state 
--
------------------------------------------------------------------------------------------
-- Dependencies : tictactoe_config_pkg.vhd
--   
------------------------------------------------------------------------------------------
-- Modifications :
-- Ver    Date        Engineer      Comments
-- 0.0    See header  FCC           Initial version from DE1_SOC_golden_top.v (Altera)
--
------------------------------------------------------------------------------------------

use work.tictactoe_config_pkg.all;

library ieee ;
  use ieee.std_logic_1164.all ;
  use ieee.numeric_std.all ;

entity DE1_SOC_golden_top is
  port (
    CLOCK_50    : in    std_logic;
    TD_CLK27    : in    std_logic;
    GPIO_0      : inout std_logic_vector(35 downto 0);
    GPIO_1      : inout std_logic_vector(35 downto 0);
    HEX0        : out   std_logic_vector(6 downto 0);
    HEX1        : out   std_logic_vector(6 downto 0);
    HEX2        : out   std_logic_vector(6 downto 0);
    HEX3        : out   std_logic_vector(6 downto 0);
    HEX4        : out   std_logic_vector(6 downto 0);
    HEX5        : out   std_logic_vector(6 downto 0);
    KEY         : in    std_logic_vector(3 downto 0);
    LEDR        : out   std_logic_vector(9 downto 0);
    SW          : in    std_logic_vector(9 downto 0);
    VGA_B       : out   std_logic_vector(7 downto 0);
    VGA_BLANK_N : out   std_logic;
    VGA_CLK     : out   std_logic;
    VGA_G       : out   std_logic_vector(7 downto 0);
    VGA_HS      : out   std_logic;
    VGA_R       : out   std_logic_vector(7 downto 0);
    VGA_SYNC_N  : out   std_logic;
    VGA_VS      : out   std_logic
    --DRAM_ADDR   : out   std_logic_vector(12 downto 0);
    --DRAM_BA     : out   std_logic_vector(1 downto 0);
    --DRAM_CAS_N  : out   std_logic;
    --DRAM_CKE    : out   std_logic;
    --DRAM_CLK    : out   std_logic;
    --DRAM_CS_N   : out   std_logic;
    --DRAM_DQ     : inout std_logic_vector(15 downto 0);
    --DRAM_LDQM   : out   std_logic;
    --DRAM_RAS_N  : out   std_logic;
    --DRAM_UDQM   : out   std_logic;
    --DRAM_WE_N   : out   std_logic 
  ) ;
end entity ; -- DE1_SOC_golden_top

architecture rtl of DE1_SOC_golden_top is

  component VGA_Ctrl is
  port(
    iRed       : in  std_logic_vector( 7 downto 0);
    iGreen     : in  std_logic_vector( 7 downto 0);
    iBlue      : in  std_logic_vector( 7 downto 0);
    oAddress   : out std_logic_vector(21 downto 0);
    oCurrent_X : out std_logic_vector(10 downto 0);
    oCurrent_Y : out std_logic_vector(10 downto 0);
    oRequest   : out std_logic;
    oVGA_R     : out std_logic_vector( 7 downto 0);
    oVGA_G     : out std_logic_vector( 7 downto 0);
    oVGA_B     : out std_logic_vector( 7 downto 0);
    oVGA_HS    : out std_logic;
    oVGA_VS    : out std_logic;
    oVGA_SYNC  : out std_logic;
    oVGA_BLANK : out std_logic;
    oVGA_CLOCK : out std_logic;
    iCLK       : in  std_logic;
    iRST_N     : in  std_logic);
  end component;


  -- SPI I/O wrapping
  alias sclk_line_a           : std_logic is GPIO_1(0); -- /!\ Need pull-up
  alias mosi_line_a           : std_logic is GPIO_1(1); -- /!\ Need pull-up
  alias miso_line_a           : std_logic is GPIO_1(2); -- /!\ Need pull-up
  alias n_ss_line_a           : std_logic is GPIO_1(3); -- /!\ Need pull-up
  alias master_request_line_a : std_logic is GPIO_1(4); -- /!\ Need pull-up

  -- Control/status I/O wrapping
  alias master_request_a : std_logic is SW(9);
  alias master_led_a     : std_logic is LEDR(9);
  alias ss_led_a         : std_logic is LEDR(8);

  -- User I/O wrapping
  alias user_n_rst_a     : std_logic is KEY(3);
  alias user_n_keys_a    : std_logic_vector(2 downto 0) is KEY(2 downto 0);
  alias user_switchs_a   : std_logic_vector(7 downto 0) is SW(7 downto 0);
  alias user_leds_a      : std_logic_vector(7 downto 0) is LEDR(7 downto 0);
  alias user_n_sevsegl_a : std_logic_vector(6 downto 0) is HEX2;
  alias user_n_sevsegc_a : std_logic_vector(6 downto 0) is HEX1;
  alias user_n_sevsegr_a : std_logic_vector(6 downto 0) is HEX0;

  -- SPI user signals
  signal master_en_s : std_logic;
  signal mst_sclk_s  : std_logic;
  signal mst_mosi_s  : std_logic;
  signal mst_miso_s  : std_logic;
  signal mst_n_ss_s  : std_logic;
  signal slv_sclk_s  : std_logic;
  signal slv_mosi_s  : std_logic;
  signal slv_miso_s  : std_logic;
  signal slv_n_ss_s  : std_logic;

  -- user signals
  signal user_rst_s     : std_logic;
  signal user_keys_s    : std_logic_vector( 2 downto 0);
  signal user_sevsegl_s : std_logic_vector( 6 downto 0);
  signal user_sevsegc_s : std_logic_vector( 6 downto 0);
  signal user_sevsegr_s : std_logic_vector( 6 downto 0);
  signal vga_r_s        : std_logic_vector( 7 downto 0);
  signal vga_g_s        : std_logic_vector( 7 downto 0);
  signal vga_b_s        : std_logic_vector( 7 downto 0);
  signal vga_x_s        : std_logic_vector(10 downto 0);
  signal vga_y_s        : std_logic_vector(10 downto 0);

  -- game state
  signal game_state_s : game_state_t;
begin

  -- not used
  GPIO_0              <= (others => 'Z');
  GPIO_1(35 downto 4) <= (others => 'Z');
  HEX5                <= (others => '1');
  HEX4                <= (others => '1');
  HEX3                <= (others => '1');

  ----------------------
  --       SPI 
  ----------------------

  -- master_request_line_a is a wired AND between the FPGAs master_request_a
  -- if master_request_line_a='1' both FPGAs are asking to be master 
  master_request_line_a <= 'Z' when master_request_a='1' else '0'; -- pull-up 

  -- we are master only if we ask for it and the other FPGA don't
  master_en_s  <= '1' when (master_request_a='1' AND master_request_line_a='0') else '0';
  master_led_a <= master_en_s;

  -- SCLK line
  sclk_line_a <= '0' when (mst_sclk_s='0'  AND master_en_s='1') else 'Z';
  slv_sclk_s  <= '0' when (sclk_line_a='0' AND master_en_s='0') else '1';

  -- MOSI line
  mosi_line_a <= '0' when (mst_mosi_s='0'  AND master_en_s='1') else 'Z';
  slv_mosi_s  <= '0' when (mosi_line_a='0' AND master_en_s='0') else '1';

  -- MISO line
  miso_line_a <= '0' when (slv_miso_s='0'  AND master_en_s='0') else 'Z';
  mst_miso_s  <= '0' when (miso_line_a='0' AND master_en_s='1') else '1';

  -- nSS line
  n_ss_line_a <= '0' when (mst_n_ss_s='0'  AND master_en_s='1') else 'Z';
  slv_n_ss_s  <= '0' when (n_ss_line_a='0' AND master_en_s='0') else '1';
  ss_led_a    <= not slv_n_ss_s;
   

  ----------------------
  --       User 
  ----------------------
   user_rst_s       <= not user_n_rst_a;
   user_keys_s      <= not user_n_keys_a;
   user_n_sevsegl_a <= not user_sevsegl_s;
   user_n_sevsegc_a <= not user_sevsegc_s;
   user_n_sevsegr_a <= not user_sevsegr_s;

  user_top : entity work.tictactoe_top
  port map(
    clk_i        => TD_CLK27,
    rst_i        => user_rst_s,
    keys_i       => user_keys_s,
    switchs_i    => user_switchs_a,
    game_state_o => game_state_s,
    master_en_i  => master_en_s,
    sclk_o       => mst_sclk_s,
    mosi_o       => mst_mosi_s,
    miso_i       => mst_miso_s,
    n_ss_o       => mst_n_ss_s,
    sclk_i       => slv_sclk_s,
    mosi_i       => slv_mosi_s,
    miso_o       => slv_miso_s,
    n_ss_i       => slv_n_ss_s
  );

  game_display : entity work.game_display
  port  map(
      clk_i           => TD_CLK27,
      rst_i           => user_rst_s,
      game_state_i    => game_state_s,
      leds_o          => user_leds_a,
      vga_red_o       => vga_r_s,
      vga_green_o     => vga_g_s,
      vga_blue_o      => vga_b_s,
      vga_current_x_i => vga_x_s,
      vga_current_y_i => vga_y_s,
      sevsegl_o       => user_sevsegl_s,
      sevsegc_o       => user_sevsegc_s,
      sevsegr_o       => user_sevsegr_s
  ) ; 

  VGA_Ctrl_inst : VGA_Ctrl
  port map(
    -- Host Side
    iRed       => vga_r_s,
    iGreen     => vga_g_s,
    iBlue      => vga_b_s,
    oCurrent_X => vga_x_s,
    oCurrent_Y => vga_y_s,
    oAddress   => open,
    oRequest   => open,
    -- VGA Side
    oVGA_R     => VGA_R,
    oVGA_G     => VGA_G,
    oVGA_B     => VGA_B,
    oVGA_HS    => VGA_HS,
    oVGA_VS    => VGA_VS,
    oVGA_SYNC  => VGA_SYNC_N,
    oVGA_BLANK => VGA_BLANK_N,
    oVGA_CLOCK => VGA_CLK,
    -- Control Signals
    iCLK       => TD_CLK27,
    iRST_N     => user_n_rst_a 
  );


end architecture ; -- rtl