------------------------------------------------------------------------------------------
-- HEIG-VD ///////////////////////////////////////////////////////////////////////////////
-- Haute Ecole d'Ingenerie et de Gestion du Canton de Vaud
-- School of Business and Engineering in Canton de Vaud
------------------------------------------------------------------------------------------
-- REDS Institute ////////////////////////////////////////////////////////////////////////
-- Reconfigurable Embedded Digital Systems
------------------------------------------------------------------------------------------
--
-- File                 : game_display.vhd
-- Author               : Flavio Capitao (FCC)
-- Date                 : 20.04.2017
--
-- Context              : Tic-Tac-Toe game
--
------------------------------------------------------------------------------------------
-- Description : Display game informations : LEDs, 7segments and VGA
--
--           /!\ Screen symbols are verry high ressources consuming => low fmax,
--               this can be solved by pre-generate symbols at synthesis...
--   
------------------------------------------------------------------------------------------
-- Dependencies : log_pkg.vhd
--                tictactoe_config_pkg.vhd
--   
------------------------------------------------------------------------------------------
-- Modifications :
-- Ver    Date        Engineer      Comments
-- 0.0    See header  FCC           Initial version.
--
------------------------------------------------------------------------------------------

use work.tictactoe_config_pkg.all;
use work.log_pkg.all;

library ieee ;
    use ieee.std_logic_1164.all ;
    use ieee.numeric_std.all ;

entity game_display is
  port (
    clk_i           : in  std_logic;
    rst_i           : in  std_logic;
    game_state_i    : in  game_state_t;
    leds_o          : out std_logic_vector( 7 downto 0);
    sevsegl_o       : out std_logic_vector( 6 downto 0);
    sevsegc_o       : out std_logic_vector( 6 downto 0);
    sevsegr_o       : out std_logic_vector( 6 downto 0);
    vga_red_o       : out std_logic_vector( 7 downto 0);
    vga_green_o     : out std_logic_vector( 7 downto 0);
    vga_blue_o      : out std_logic_vector( 7 downto 0);
    vga_current_x_i : in  std_logic_vector(10 downto 0);
    vga_current_y_i : in  std_logic_vector(10 downto 0)
  ) ;
end entity ; -- game_display

architecture behave of game_display is

    signal player1_led_s     : std_logic;
    signal player2_led_s     : std_logic;
    signal sel_led_s         : std_logic;

    -- Clock divider    
    constant DIV_US        : integer := 200000; -- us (half period of div_s)
    constant DIV_COUNT     : integer := CLK_FREQ*DIV_US; -- iterations=freq*timerTime
    signal   div_s         : std_logic;
    signal   div_counter_s : unsigned(ilogup(DIV_COUNT-1) downto 0);

    -- PWM
    constant PWM_PERIOD_US : integer := 200; -- us
    constant PWM_DUTY      : integer := 10; -- %
    constant PWM_COUNT     : integer := CLK_FREQ*PWM_PERIOD_US; -- iterations=freq*timerTime
    constant PWM_THR       : integer := PWM_COUNT*PWM_DUTY/100;
    signal   pwm_s         : std_logic;
    signal   pwm_counter_s : unsigned(ilogup(PWM_COUNT-1) downto 0); 

    -- VGA
    constant SCREEN_WIDTH         : integer := 640;
    constant SCREEN_HEIGHT        : integer := 480;
    constant ZONE_SIZE            : integer := 128;
    constant DRAW_SIZE            : integer := ZONE_SIZE*8/10;
    constant DRAW_LINE_WIDTH      : integer := 4;
    constant DRAW_LINE_DIAG_WIDTH : integer := 6; -- best : DRAW_LINE_WIDTH*sqrt(2)
    constant GAME_ZONE_SIZE       : integer := ZONE_SIZE*3+DRAW_LINE_WIDTH*2;

    constant BACKGROUND_COLOR   : std_logic_vector(23 downto 0) := x"FFFFFF";

    constant PLAYER1_COLOR      : std_logic_vector(23 downto 0) := x"FF0000";
    constant PLAYER1_WIN_COLOR  : std_logic_vector(23 downto 0) := x"FF0000";
    constant PLAYER1_LOSE_COLOR : std_logic_vector(23 downto 0) := x"800000";

    constant PLAYER2_COLOR      : std_logic_vector(23 downto 0) := x"0000FF";
    constant PLAYER2_WIN_COLOR  : std_logic_vector(23 downto 0) := x"0000FF";
    constant PLAYER2_LOSE_COLOR : std_logic_vector(23 downto 0) := x"000080";

    constant WIN_LINE_COLOR     : std_logic_vector(23 downto 0) := x"00FF00";
    constant LINE_COLOR         : std_logic_vector(23 downto 0) := x"000000";

    constant DEMO_COLOR         : std_logic_vector(23 downto 0) := x"FF0000";

    signal   color_out_s        : std_logic_vector(23 downto 0); 

    -- Demo draw
    constant DEMO_EN     : boolean := false;
    constant DEMO_WIDTH  : integer := 28;
    constant DEMO_HEIGHT : integer := 11;
    type demo_draw_t is array (0 to DEMO_HEIGHT-1) of std_logic_vector(0 to DEMO_WIDTH-1);
    constant DEMO : demo_draw_t := (
        "00"&"00000"&"00"&"00000"&"00"&"00000"&"00"&"00000",
        "00"&"00000"&"00"&"00000"&"00"&"00000"&"00"&"00000",
        "00"&"11100"&"00"&"11111"&"00"&"11111"&"00"&"11111",
        "00"&"10010"&"00"&"10000"&"00"&"10101"&"00"&"10001",
        "00"&"10010"&"00"&"10000"&"00"&"10101"&"00"&"10001",
        "00"&"10001"&"00"&"10000"&"00"&"10101"&"00"&"10001",
        "00"&"10001"&"00"&"11111"&"00"&"10101"&"00"&"10001",
        "00"&"10001"&"00"&"10000"&"00"&"10101"&"00"&"10001",
        "00"&"10010"&"00"&"10000"&"00"&"10101"&"00"&"10001",
        "00"&"10010"&"00"&"10000"&"00"&"10101"&"00"&"10001",
        "00"&"11100"&"00"&"11111"&"00"&"10101"&"00"&"11111"
    );

begin

    -- clock divider for selection
    div_gen : process( clk_i, rst_i )
    begin
        if rst_i = '1' then
            div_counter_s <= to_unsigned(DIV_COUNT-1,div_counter_s'length);
            div_s   <= '0'; 
        elsif rising_edge(clk_i) then
            div_counter_s <= div_counter_s-1;
            if div_counter_s=0 then
                div_counter_s <= to_unsigned(DIV_COUNT-1,div_counter_s'length);
                div_s         <= not div_s;
            end if ;
        end if ;        
    end process ; -- clk_div

   ---------------------------------------------------------------------------------
    --                                    VGA
    --------------------------------------------------------------------------------
    vga_red_o   <= color_out_s(23 downto 16);
    vga_green_o <= color_out_s(15 downto  8);
    vga_blue_o  <= color_out_s( 7 downto  0);

    vga_display : process( clk_i,rst_i )
        ----------------
        function circle (x,y : integer) return boolean is
            constant RADIUS  : integer := DRAW_SIZE/2;
            variable xc,yc,a : integer;
        begin
            xc:=x-ZONE_SIZE/2; -- center point
            yc:=y-ZONE_SIZE/2;
            a:=xc*xc+yc*yc;       -- current distance from center
            return (a<(RADIUS**2)) AND (a>(RADIUS-DRAW_LINE_WIDTH)**2);
        end circle;
        ----------------
        function point (x,y : integer) return boolean is
            constant RADIUS  : integer := DRAW_SIZE/8;
            variable xc,yc,a : integer;
        begin
            xc:=x-ZONE_SIZE/2; -- center point
            yc:=y-ZONE_SIZE/2;
            a:=xc*xc+yc*yc;    -- current distance from center
            return a<(RADIUS**2);
        end point;
        ----------------
        function cross (x,y : integer) return boolean is
            constant MARGIN : integer := (ZONE_SIZE-DRAW_SIZE)/2;
        begin
            return (((y-DRAW_LINE_DIAG_WIDTH/2) < x AND x < (y+DRAW_LINE_DIAG_WIDTH/2)) OR -- line \
                    ((ZONE_SIZE-y-DRAW_LINE_DIAG_WIDTH/2) < x AND x < (ZONE_SIZE-y+DRAW_LINE_DIAG_WIDTH/2)) ) AND -- line /
                    x>MARGIN AND x<(ZONE_SIZE-MARGIN) ; -- margin
        end cross;
        ----------------
        variable draw_circle,draw_cross,draw_point,draw_win_line,draw_line,draw_demo : boolean;
        variable x,y,zpx,zpy,zn,offset : integer;
    begin
        if rst_i='1' then
            color_out_s <= (others => '0');
        elsif rising_edge(clk_i) then

            -- Init
            draw_circle   := false;
            draw_cross    := false;
            draw_point    := false;
            draw_win_line := false;
            draw_line     := false;
            draw_demo     := false;
            x := to_integer(unsigned(vga_current_x_i));
            y := to_integer(unsigned(vga_current_y_i));

            -- Draw demo
            if x < DEMO_WIDTH AND y < DEMO_HEIGHT AND DEMO(y)(x)='1' then
                draw_demo := true; 
            end if;

            -- Center on screen
            x := x- SCREEN_WIDTH/2+DRAW_LINE_WIDTH+ZONE_SIZE*3/2;
            y := y-SCREEN_HEIGHT/2+DRAW_LINE_WIDTH+ZONE_SIZE*3/2;

            -------------------------- Draw lines --------------------------

            if 0 < x AND 0 < y AND x < GAME_ZONE_SIZE AND y < GAME_ZONE_SIZE AND -- game zone
              (( (ZONE_SIZE)                   < x AND x < (ZONE_SIZE  +DRAW_LINE_WIDTH)   ) OR    -- line vertical left
               ( (ZONE_SIZE*2+DRAW_LINE_WIDTH) < x AND x < (ZONE_SIZE*2+DRAW_LINE_WIDTH*2) ) OR    -- line vertical rigth
               ( (ZONE_SIZE)                   < y AND y < (ZONE_SIZE  +DRAW_LINE_WIDTH)   ) OR    -- line horizontal up
               ( (ZONE_SIZE*2+DRAW_LINE_WIDTH) < y AND y < (ZONE_SIZE*2+DRAW_LINE_WIDTH*2) ) )then -- line horizontal down
                draw_line := true;
            end if ;

            -------------------------- Draw players marks --------------------------

            zone_y : for zy in 0 to 2 loop
                zone_x : for zx in 0 to 2 loop
                    zn  := zx+3*zy;
                    zpx := x-(ZONE_SIZE*zx+DRAW_LINE_WIDTH*zx);
                    zpy := y-(ZONE_SIZE*zy+DRAW_LINE_WIDTH*zy);
                    if 0 < zpx AND zpx < ZONE_SIZE AND 0 < zpy AND zpy < ZONE_SIZE then

                        if game_state_i.player1_reg(zn)='1' then
                            draw_cross  := cross(zpx,zpy);
                        end if ;

                        if game_state_i.player2_reg(zn)='1' then
                            draw_circle := circle(zpx,zpy);
                        end if ;

                        if game_state_i.sel_reg(zn)='1' AND div_s='1' AND game_state_i.curr_player_win='0' AND game_state_i.draw_match='0' then
                            draw_point  := point(zpx,zpy);
                        end if ;

                    end if ;
                end loop ; -- zone_x
            end loop ; -- zone_y

            -------------------------- Draw win line --------------------------

            horiz_verti_win_lines : for i in 0 to 2 loop
                offset := ZONE_SIZE/2+i*ZONE_SIZE+i*DRAW_LINE_WIDTH;

                -- row 1,2,3
                if game_state_i.win_line(i)='1' then
                    if 0 < x AND 0 < y AND x < GAME_ZONE_SIZE AND y < GAME_ZONE_SIZE AND -- game zone
                      (offset-DRAW_LINE_WIDTH/2) < y AND y < (offset+DRAW_LINE_WIDTH/2) then 
                        draw_win_line := true;
                    end if ;
                end if ;

                -- col 1,2,3
                if game_state_i.win_line(i+3)='1' then 
                    if 0 < x AND 0 < y AND x < GAME_ZONE_SIZE AND y < GAME_ZONE_SIZE AND -- game zone
                      (offset-DRAW_LINE_WIDTH/2) < x AND x < (offset+DRAW_LINE_WIDTH/2) then 
                        draw_win_line := true;
                    end if ;
                end if ;
            end loop ; -- horiz_verti_win_lines

            -- diag \
            if game_state_i.win_line(6)='1' then
                if 0 < x AND 0 < y AND x < GAME_ZONE_SIZE AND y < GAME_ZONE_SIZE AND -- game zone
                  ( ((y-DRAW_LINE_WIDTH/2) < x AND x < (y+DRAW_LINE_WIDTH/2)) ) then 
                    draw_win_line := true;
                end if ;
            end if ; 

            -- diag /
            if game_state_i.win_line(7)='1' then
                if 0 < x AND 0 < y AND x < GAME_ZONE_SIZE AND y < GAME_ZONE_SIZE AND -- game zone
                  ((GAME_ZONE_SIZE-y-DRAW_LINE_DIAG_WIDTH/2) < x AND x < (GAME_ZONE_SIZE-y+DRAW_LINE_DIAG_WIDTH/2)) then
                    draw_win_line := true;
                end if ;
            end if ;

            -------------------------- Draw colors --------------------------

            color_out_s <= BACKGROUND_COLOR;

            -- lines 
            if draw_line then
                color_out_s   <= LINE_COLOR; 
            end if;

            -- player 1
            if draw_cross OR (draw_point AND game_state_i.curr_player='0') then
                color_out_s   <= PLAYER1_COLOR;
                if game_state_i.draw_match='1' then
                    color_out_s <= PLAYER1_LOSE_COLOR;
                end if ;
                if game_state_i.curr_player_win='1' then
                    if game_state_i.curr_player='0' then
                        color_out_s <= PLAYER1_WIN_COLOR;
                    else
                        color_out_s <= PLAYER1_LOSE_COLOR;
                    end if ;
                end if ;
            end if;

            -- player 2
            if draw_circle OR (draw_point AND game_state_i.curr_player='1') then
                color_out_s   <= PLAYER2_COLOR;
                if game_state_i.draw_match='1' then
                    color_out_s <= PLAYER2_LOSE_COLOR;
                end if ;
                if game_state_i.curr_player_win='1' then
                    if game_state_i.curr_player='1' then
                        color_out_s <= PLAYER2_WIN_COLOR;
                    else
                        color_out_s <= PLAYER2_LOSE_COLOR;
                    end if ;
                end if ;
            end if;  

            -- win line 
            if draw_win_line then
                color_out_s   <= WIN_LINE_COLOR; 
            end if;

            -- demo
            if DEMO_EN AND draw_demo then
                color_out_s <= DEMO_COLOR; 
            end if;

        end if ;
    end process ; -- vga_display


    --------------------------------------------------------------------------------
    --                                    LEDs
    --------------------------------------------------------------------------------

    player1_led_s <= '1';
    player2_led_s <= pwm_s;
    sel_led_s <= div_s and player1_led_s and not (game_state_i.curr_player_win or game_state_i.draw_match) 
                when game_state_i.curr_player='0' else
             div_s and player2_led_s and not (game_state_i.curr_player_win or game_state_i.draw_match);

    -- PWM gen ( for user 2 leds)
    pwm_gen : process( clk_i, rst_i )
    begin
        if rst_i = '1' then
            pwm_counter_s <= to_unsigned(0,pwm_counter_s'length);
            pwm_s   <= '0'; 
        elsif rising_edge(clk_i) then

            if pwm_counter_s<PWM_THR then
                pwm_s   <= '1';
            else
                pwm_s   <= '0';
            end if ;

            if pwm_counter_s=PWM_COUNT-1 then
                pwm_counter_s <= to_unsigned(0,pwm_counter_s'length);
            else
                pwm_counter_s <= pwm_counter_s+1;
            end if ;
        end if ;        
    end process ;

    leds_o(0) <= game_state_i.curr_player_win;
    leds_o(1) <= game_state_i.draw_match;
    leds_o(2) <= game_state_i.curr_player;
    leds_o(3) <= '0';
    leds_o(4) <= player1_led_s when game_state_i.curr_player='0' else player2_led_s;
    leds_o(5) <= '0';
    leds_o(6) <= '0';
    leds_o(7) <= '1' when DEMO_EN else '0';


    --------------------------------------------------------------------------------
    --                                  7 segments
    --------------------------------------------------------------------------------

    sevsegl_o(1) <= '0';
    sevsegl_o(2) <= '0';
    sevsegl_o(4) <= '0';
    sevsegl_o(5) <= '0';

    sevsegc_o(1) <= '0';
    sevsegc_o(2) <= '0';
    sevsegc_o(4) <= '0';
    sevsegc_o(5) <= '0';

    sevsegr_o(1) <= '0';
    sevsegr_o(2) <= '0';
    sevsegr_o(4) <= '0';
    sevsegr_o(5) <= '0';


    sevsegl_o(0) <= player1_led_s when game_state_i.player1_reg(0)='1' else
                    player2_led_s when game_state_i.player2_reg(0)='1' else
                    sel_led_s     when     game_state_i.sel_reg(0)='1'  
                    else '0';

    sevsegc_o(0) <= player1_led_s when game_state_i.player1_reg(1)='1' else
                    player2_led_s when game_state_i.player2_reg(1)='1' else
                    sel_led_s     when     game_state_i.sel_reg(1)='1'  
                    else '0';

    sevsegr_o(0) <= player1_led_s when game_state_i.player1_reg(2)='1' else
                    player2_led_s when game_state_i.player2_reg(2)='1' else
                    sel_led_s     when     game_state_i.sel_reg(2)='1'  
                    else '0';

    sevsegl_o(6) <= player1_led_s when game_state_i.player1_reg(3)='1' else
                    player2_led_s when game_state_i.player2_reg(3)='1' else
                    sel_led_s     when     game_state_i.sel_reg(3)='1'  
                    else '0';

    sevsegc_o(6) <= player1_led_s when game_state_i.player1_reg(4)='1' else
                    player2_led_s when game_state_i.player2_reg(4)='1' else
                    sel_led_s     when     game_state_i.sel_reg(4)='1'  
                    else '0';

    sevsegr_o(6) <= player1_led_s when game_state_i.player1_reg(5)='1' else
                    player2_led_s when game_state_i.player2_reg(5)='1' else
                    sel_led_s     when     game_state_i.sel_reg(5)='1'  
                    else '0';

    sevsegl_o(3) <= player1_led_s when game_state_i.player1_reg(6)='1' else
                    player2_led_s when game_state_i.player2_reg(6)='1' else
                    sel_led_s     when     game_state_i.sel_reg(6)='1'  
                    else '0';

    sevsegc_o(3) <= player1_led_s when game_state_i.player1_reg(7)='1' else
                    player2_led_s when game_state_i.player2_reg(7)='1' else
                    sel_led_s     when     game_state_i.sel_reg(7)='1'  
                    else '0';

    sevsegr_o(3) <= player1_led_s when game_state_i.player1_reg(8)='1' else
                    player2_led_s when game_state_i.player2_reg(8)='1' else
                    sel_led_s     when     game_state_i.sel_reg(8)='1'  
                    else '0';


end architecture ; -- behave
